/*
** prpn.c: A RPN calculator
**
** Copyright (c) 2003 Peter Eriksson <pen@lysator.liu.se>
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <time.h>
#include <limits.h>


char *argv0 = "prpn";
double *stack = NULL;
int stacksize = 0;
int stackpos = 0;
int sum_f = 0;
int nan_f = 0;
char t_suffix = 0;
int dotprec = -1;

#define STACK(n) stack[stackpos-(n)-1]

struct siunit {
    char suffix;
    int exp;
} suffixtab[] = {
    { 'Y', 24 },
    { 'Z', 21 },
    { 'E', 18 },
    { 'P', 15 },
    { 'T', 12 },
    { 'G',  9 },
    { 'M',  6 },
    { 'k',  3 },
    { 'h',  2 },
#if 0
    { 'da', 1 },
#endif
    { ' ',  0 },
    { 'd', -1 },
    { 'c', -2 },
    { 'm', -3 },
    { '', -6 },
    { 'n', -9 },
    { 'p', -12 },
    { 'f', -15 },
    { 'z', -21 },
    { 'y', -24 },
    { 0, 0 }
};


void v_help(void);

void v_plus(void);
void v_minus(void);
void v_multiply(void);
void v_divide(void);
void v_percent(void);
void v_pchange(void);
void v_pt(void);
void v_pi(void);
void v_e(void);
void v_neg(void);
void v_rand(void);
void v_rdz(void);
void v_maxr(void);
void v_minr(void);
void v_abs(void);
void v_sign(void);
void v_ip(void);
void v_fp(void);
void v_max(void);
void v_min(void);
void v_mod(void);

void v_ln(void);
void v_log(void);
void v_cos(void);
void v_cosh(void);
void v_acos(void);
void v_asin(void);
void v_atan(void);
void v_ceil(void);
void v_exp(void);
void v_floor(void);
void v_pow(void);
void v_sin(void);
void v_sinh(void);
void v_sqrt(void);
void v_tan(void);
void v_tanh(void);

void v_isnan(void);

void v_drop(void);
void v_dup(void);
void v_depth(void);
void v_swap(void);
void v_clear(void);
void v_roll(void);
void v_over(void);
void v_dup2(void);
void v_rot(void);
void v_rolld(void);
void v_pick(void);
void v_dupn(void);
void v_dropn(void);


void v_dot(void);
void v_p(void);
void v_exit(void);
void v_prec(void);



struct token {
    char *name;
    void (*call)(void);
} tokentab[] = {
    { "help", v_help },
    { "+", v_plus },
    { "-", v_minus },
    { "*", v_multiply },
    { "/", v_divide },
    { "%", v_percent },
    { "%ch", v_pchange },
    { "%t", v_pt },
    { "neg", v_neg },
    { "rand", v_rand },
    { "rdz", v_rdz },
    { "maxr", v_maxr },
    { "minr", v_minr },
    { "abs", v_abs },
    { "sign", v_sign },
    { "ip", v_ip },
    { "fp", v_fp },
    { "max", v_max },
    { "min", v_min },
    { "mod", v_mod },
    
    { "ln", v_ln },
    { "log", v_log },
    { "cos", v_cos },
    { "cosh", v_cosh },
    { "acos", v_acos },
    { "asin", v_asin },
    { "atan", v_atan },
    { "ceil", v_ceil },
    { "exp", v_exp },
    { "floor", v_floor },
    { "pow", v_pow },
    { "sin", v_sin },
    { "sinh", v_sinh },
    { "sqrt", v_sqrt },
    { "tan", v_tan },
    { "tanh", v_tanh },
    { "isnan", v_isnan },
    { "e", v_e },
    { "pi", v_pi },
    
    { "drop", v_drop },
    { "dup", v_dup },
    { "depth", v_depth },
    { "swap", v_swap },
    { "clear", v_clear },
    { "roll", v_roll },
    { "over", v_over },
    { "dup2", v_dup2 },
    { "rot", v_rot },
    { "rolld", v_rolld },
    { "pick", v_pick },
    { "dupn", v_dupn },
    { "dropn", v_dropn },
   
    { ".", v_dot },
    { "p", v_p },
    
    { "exit", v_exit },
    { "prec", v_prec },
    { NULL, NULL }
};



void
v_help(void)
{
    int i, j;

    puts("Available commands/operators:");
    for (i = 0; tokentab[i].name; i++)
	printf("%s%s", tokentab[i].name, (!((i+1)&7) || !tokentab[i+1].name) ? "\n" : " ");
}



void
stack_push(double val)
{
    if (stackpos >= stacksize)
    {
	stacksize += 256;
	if (!stack)
	    stack = malloc(sizeof(double)*stacksize);
	else
	    stack = realloc(stack, sizeof(double)*stacksize);
	if (!stack)
	{
	    fprintf(stderr, "%s: stack_push: allocation failed: %s\n",
		    argv0, strerror(errno));
	    exit(1);
	}
    }

    stack[stackpos++] = val;
}

double
stack_pop(void)
{
    double val;

    
    if (stackpos > 0)
    {
	val = stack[--stackpos];
	return val;
    }

    fputs("> ", stdout);
    fflush(stdout);
    if (scanf("%lf", &val) == 1)
	return val;

    fprintf(stderr, "%s: POP: %s\n", argv0, strerror(errno));
    exit(1);
}




int
suffix2exp(int suffix, int *exp)
{
    int i;
    
    if (!suffix)
    {
	*exp = 0;
	return 0;
    }
    
    for (i = 0; suffixtab[i].suffix && suffix != suffixtab[i].suffix; i++)
	;
    
    if (suffixtab[i].suffix)
    {
	*exp = suffixtab[i].exp;
	return i;
    }
    
    return -1;
}


void
print_value(double val)
{
    int j, ev;


    if (t_suffix < 0)
    {
	if (dotprec >= 0)
	    printf("%.*lg", dotprec, val);
	else
	    printf("%lg", val);
    }
    else
    {
	ev = log10(fabs(val));

	for (j = 0; suffixtab[j].suffix &&
		 ((suffixtab[j].exp < 3 && suffixtab[j].exp > -3 && suffixtab[j].exp != 0) ||
		 ev < suffixtab[j].exp); j++)
	    ;

	if (suffixtab[j].suffix)
	    val /= pow(10.0, suffixtab[j].exp);

	if (dotprec >= 0)
	{
	    printf("%.*lf", dotprec, val);
	    if (suffixtab[j].suffix)
		putchar(suffixtab[j].suffix);
	}
	else
	{
	    printf("%lf", val);
	    if (j > 0)
		putchar(suffixtab[j].suffix);
	}
    }
    
    putchar('\n');
}


void
v_plus(void)
{
    double v1, v2;
    
    v1 = stack_pop();
    v2 = stack_pop();
    stack_push(v2+v1);
}
    
void
v_minus(void)
{
    double v1, v2;
    
    v1 = stack_pop();
    v2 = stack_pop();
    stack_push(v2-v1);
}

void
v_multiply(void)
{
    double v1, v2;
    
    v1 = stack_pop();
    v2 = stack_pop();
    stack_push(v2*v1);
}

void
v_divide(void)
{
    double v1, v2;
    
    v1 = stack_pop();
    v2 = stack_pop();
    stack_push(v2/v1);
}

void
v_percent(void)
{
    double v1, v2;
    
    v1 = stack_pop();
    v2 = stack_pop();
    stack_push(v2*v1/100);
}

void
v_pchange(void)
{
    double v1, v2;
    
    v1 = stack_pop();
    v2 = stack_pop();
    stack_push(100*(v1-v2)/v2);
}

void
v_pt(void)
{
    double x, y;
    
    y = stack_pop();
    x = stack_pop();
    
    stack_push(100*y/x);
}


void
v_neg(void)
{
    stack_push(-stack_pop());
}


void
v_rand(void)
{
    stack_push(drand48());
}

void
v_rdz(void)
{
    int seed = (long) stack_pop();
    
    srand48(seed ? seed : time(NULL));
}

void
v_maxr(void)
{
    stack_push(DBL_MAX);
}

void
v_minr(void)
{
    stack_push(DBL_MIN);
}

void
v_abs(void)
{
    stack_push(fabs(stack_pop()));
}


void
v_sign(void)
{
    double val = stack_pop();

    if (val < 0)
	stack_push(-1);
    else if (val > 0)
	stack_push(1);
    else
	stack_push(0);
}


void
v_ip(void)
{
    double ip;
    double val = stack_pop();

    
    if (val < 0)
	ip = ceil(val);
    else
	ip = floor(val);

    stack_push(ip);
}

void
v_fp(void)
{
    double ip;
    double val = stack_pop();

    
    if (val < 0)
	ip = ceil(val);
    else
	ip = floor(val);

    stack_push(val-ip);
}


void
v_max(void)
{
    double v1 = stack_pop();
    double v2 = stack_pop();

    if (v1 > v2)
	stack_push(v1);
    else
	stack_push(v2);
}

void
v_min(void)
{
    double v1 = stack_pop();
    double v2 = stack_pop();

    if (v1 < v2)
	stack_push(v1);
    else
	stack_push(v2);
}

#if 0
void
v_mod(void)
{
    double y = stack_pop();
    double x = stack_pop();

    if (y == 0)
    {
	fprintf(stderr, "Error: MOD: Invalid level 1 argument\n");
	return;
    }
    
    stack_push(x-y*floor(x/y));
}
#endif

void
v_ln(void)
{
    stack_push(log(stack_pop()));
}


void
v_log(void)
{
    stack_push(log10(stack_pop()));
}

void
v_cos(void)
{
    stack_push(cos(stack_pop()));
}


void
v_cosh(void)
{
    stack_push(cosh(stack_pop()));
}


void
v_acos(void)
{
    stack_push(acos(stack_pop()));
}

void
v_asin(void)
{
    stack_push(asin(stack_pop()));
}

void
v_atan(void)
{
    stack_push(atan(stack_pop()));
}


void
v_ceil(void)
{
    stack_push(ceil(stack_pop()));
}



void
v_exp(void)
{
    stack_push(exp(stack_pop()));
}




void
v_floor(void)
{
    stack_push(floor(stack_pop()));
}

void
v_mod(void)
{
    double v1 = stack_pop();
    double v2 = stack_pop();
    
    stack_push(fmod(v2, v1));
}

void
v_pow(void)
{
    double v1 = stack_pop();
    double v2 = stack_pop();
    
    stack_push(pow(v2, v1));
}

void
v_sin(void)
{
    stack_push(sin(stack_pop()));
}

void
v_sinh(void)
{
    stack_push(sinh(stack_pop()));
}

void
v_sqrt(void)
{
    stack_push(sqrt(stack_pop()));
}


void
v_tan(void)
{
    stack_push(tan(stack_pop()));
}


void
v_tanh(void)
{
    stack_push(tanh(stack_pop()));
}

void
v_isnan(void)
{
    stack_push(isnan(stack_pop()));
}


void
v_e(void)
{
    stack_push(M_E);
}


void
v_pi(void)
{
    stack_push(M_PI);
}

    
void
v_drop(void)
{
    (void) stack_pop();
}


void
v_dup(void)
{
    double v1 = stack_pop();
    stack_push(v1);
    stack_push(v1);
}


void
v_depth(void)
{
    stack_push((double) stackpos);
}


void
v_swap(void)
{
    double v1, v2;
    
    v1 = stack_pop();
    v2 = stack_pop();
    stack_push(v1);
    stack_push(v2);
}
    
void
v_clear(void)
{
    stackpos = 0;
}

void
v_roll(void)
{
    int n, i;
    double val;
    

    n = (int) stack_pop();

    if (n <= 0)
    {
	fprintf(stderr, "Error: %d ROLL: Invalid roll index\n", n);
	return;
    }
    
    if (stackpos - n < 0)
    {
	fprintf(stderr, "Error: %d ROLL: Not enough values on stack\n", n);
	return;
    }

    val = STACK(n-1);
    while (--n > 0)
	STACK(n) = STACK(n-1);
    
    STACK(0) = val;
}


void
v_over(void)
{
    double val;

    if (stackpos <= 1)
    {
	fprintf(stderr, "Error: OVER: Not enough values on stack\n");
	return;
    }
    
    stack_push(STACK(1));
}


void
v_dup2(void)
{
    double v1;
    double v2;

    if (stackpos <= 1)
    {
	fprintf(stderr, "Error: OVER: Not enough values on stack\n");
	return;
    }

    v1 = STACK(1);
    v2 = STACK(0);
    stack_push(v1);
    stack_push(v2);
}


void
v_rot(void)
{
    int i, n;
    double val;
    
    
    if (stackpos <= 2)
    {
	fprintf(stderr, "Error: ROT: Not enough values on stack\n");
	return;
    }

    n = 3;
    
    val = STACK(n-1);
    while (--n > 0)
	STACK(n) = STACK(n-1);
    
    STACK(0) = val;
}


void
v_rolld(void)
{
    int n, i;
    double val;
    

    n = (int) stack_pop();

    if (n <= 0)
    {
	fprintf(stderr, "Error: %d ROLLD: Invalid roll index\n", n);
	return;
    }
    
    if (stackpos - n < 0)
    {
	fprintf(stderr, "Error: %d ROLLD: Not enough values on stack\n", n);
	return;
    }

    val = STACK(0);
    for (i = 1; i < n; i++)
	STACK(i-1) = STACK(i);
    STACK(i-1) = val;
}


void
v_dupn(void)
{
    int n = (int) stack_pop();
    int pos, i;

    
    if (n <= 0)
    {
	fprintf(stderr, "Error: %d DUPN: Invalid count\n", n);
	return;
    }
    
    if (stackpos - n < 0)
    {
	fprintf(stderr, "Error: %d DUPN: Not enough values on stack\n", n);
	return;
    }

    pos = stackpos-n;
    for (i = 0; i < n; i++)
	stack_push(stack[pos+i]);
}

void
v_pick(void)
{
    int n;

    if (stackpos <= 0)
    {
	fprintf(stderr, "Error: %d PICK: Not enough values on stack\n", n);
	return;
    }

    n = (int) STACK(0);
    
    if (n <= 0)
    {
	fprintf(stderr, "Error: %d PICK: Invalid index\n", n);
	return;
    }
    
    if (stackpos - n <= 0)
    {
	fprintf(stderr, "Error: %d PICK: Not enough values on stack\n", n);
	return;
    }

    STACK(0) = STACK(n);
}


void
v_dropn(void)
{
    int n = (int) stack_pop();

    if (n <= 0)
    {
	fprintf(stderr, "Error: %d DROPN: Invalid drop index\n", n);
	return;
    }
    
    if (stackpos - n <= 0)
    {
	fprintf(stderr, "Error: %d DROPN: Not enough values on stack\n", n);
	return;
    }
    
    stackpos -= n;
}


void
v_dot(void)
{
    print_value(stack_pop());
}

void
v_p(void)
{
    if (stackpos <= 0)
    {
	fprintf(stderr, "Error: P: Not enough values on stack\n");
	return;
    }
    
    print_value(STACK(0));
}


void
v_exit(void)
{
    exit((int) stack_pop());
}

void
v_prec(void)
{
    
    dotprec = (int) stack_pop();
}


int
token_call(char *token)
{
    int i;

    for (i = 0; tokentab[i].name && strcasecmp(token, tokentab[i].name); ++i)
	;

    if (tokentab[i].call)
    {
	tokentab[i].call();
	return i;
    }

    return -1;
}


int
token_eval(char *token)
{
    double val;
    char suffix = 0;
    int expval = 0;
    void (*tfun)(void);
    
    
    if (sscanf(token, "%lf%c", &val, &suffix) >= 1)
    {
	if (nan_f && isnan(val))
	    val = 0;
	
	if (suffix)
	{
	    if (suffix2exp(suffix, &expval) < 0)
	    {
		fprintf(stderr, "%s: invalid suffix: %s (%d)\n",
			argv0, token, suffix);
		exit(1);
	    }
	}
	else
	    expval = 0;
	
	
	if (expval)
	    val *= pow(10.0, expval);
	
	stack_push(val);
    }

    else if (token_call(token) < 0)
    {
	fprintf(stderr, "%s: Invalid number/token: %s\n",
		argv0, token);
	return -1;
    }

    return 0;
}


char *
token_get(void)
{
    static char buf[1024];

    if (scanf("%s", buf) == 1)
	return buf;

    return NULL;
}


int
main(int argc,
     char *argv[])
{
    int i, j, len;
    double val, tval;
    char *token = NULL;
    

    if (argv[0])
	argv0 = argv[0];
    
    tval = 0;
    
    for (i = 1; i < argc && argv[i][0] == '-'; ++i)
	switch (argv[i][1])
	{
	  case '-':
	    ++i;
	    goto Values;

	  case 'n':
	    nan_f++;
	    break;
	    
	  case 's':
	    sum_f++;
	    break;

	  case 'S':
	    t_suffix = argv[i][2];
	    if (!t_suffix)
	      t_suffix = ' ';
	    else if (t_suffix == '-')
	      t_suffix = -1;
	    break;
	    
	  case 'p':
	    if (sscanf(argv[i]+2, "%u", &dotprec) != 1)
	    {
		fprintf(stderr, "%s: invalid precision: %s\n",
			argv0, argv[i]);
		exit(1);
	    }
	    break;

	  default:
	    fprintf(stderr, "%s: invalid switch: %s\n",
		    argv0, argv[i]);
	    exit(1);
	}

  Values:
    if (i < argc)
    {
	for (;i < argc; ++i)
	{
	    token = argv[i];
	    
	    token_eval(token);
	}
    }
    else
    {
	while (token = token_get())
	    token_eval(token);
    }

    for (i = 0; i < stackpos; i++)
    {
	val = stack[i];

	if (sum_f)
	{
	    tval += val;
	    if (i+1 < stackpos)
		continue;
	}
	else
	    tval = val;
    
	print_value(tval);
    }
    
    return 0;
}
